from decimal import Decimal

from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from jsonfield import JSONField

from extra_settings import fields
from extra_settings.cache import get_cached_setting, set_cached_setting
from extra_settings.utils import enforce_uppercase_setting, import_function


class Setting(models.Model):
    """
    This class describes a Setting model.
    """

    @staticmethod
    def _cache_all_settings():
        settings_qs = Setting.objects.all()
        settings_list = list(settings_qs)
        for setting_obj in settings_list:
            setting_name = setting_obj.name
            setting_value = setting_obj.value
            set_cached_setting(setting_name, setting_value)

    @staticmethod
    def _get_from_cache(name):
        return get_cached_setting(name)

    @staticmethod
    def _get_from_database(name):
        try:
            setting_obj = Setting.objects.get(name=name)
            value = setting_obj.value
            set_cached_setting(name, value)
            return value
        except Setting.DoesNotExist:
            return None

    @staticmethod
    def _get_from_settings(name):
        if settings.EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS:
            return getattr(settings, name, None)
        return None

    @classmethod
    def get(cls, name, default=None):
        val = cls._get_from_cache(name)
        if val is None:
            val = cls._get_from_database(name)
            if val is None:
                val = cls._get_from_settings(name)
                if val is None:
                    val = default
            else:
                # value found, it means that cache was not set or expired.
                # cache all other settings to avoid multiple queries
                # in pages using the template more than once.
                cls._cache_all_settings()
        if val is not None:
            set_cached_setting(name, val)
        return val

    @classmethod
    def set_defaults(cls, defaults):
        if not isinstance(defaults, list):
            raise ValueError("Setting 'defaults' must be a list of dicts items.")
        if not defaults:
            return
        for item in defaults:
            if not isinstance(item, dict):
                raise ValueError("Setting 'defaults' item must be a dict.")
            try:
                name = item["name"]
                value_type = item["type"]
                value_type = value_type.replace("Setting.TYPE_", "").lower()
                value = item["value"]
            except KeyError as error:
                raise ValueError(
                    "Setting 'defaults' item must contain "
                    "'name', 'type' and 'value' keys."
                ) from error
            validator = item.get("validator", None)
            description = item.get("description", "")
            setting_obj, setting_created = cls.objects.get_or_create(
                name=name,
                defaults={
                    "value_type": value_type,
                    "description": description,
                    "validator": validator,
                },
            )
            if setting_created:
                setting_obj.value = value
                setting_obj.save()

    @classmethod
    def set_defaults_from_settings(cls, *args, **kwargs):
        cls.set_defaults(settings.EXTRA_SETTINGS_DEFAULTS)

    TYPE_BOOL = "bool"
    # TYPE_COLOR = "color" # TODO
    TYPE_DATE = "date"
    TYPE_DATETIME = "datetime"
    TYPE_DECIMAL = "decimal"
    TYPE_DURATION = "duration"
    TYPE_EMAIL = "email"
    TYPE_FILE = "file"
    TYPE_FLOAT = "float"
    # TYPE_HTML = "html"
    TYPE_IMAGE = "image"
    TYPE_INT = "int"
    TYPE_JSON = "json"
    TYPE_STRING = "string"
    TYPE_TEXT = "text"
    TYPE_TIME = "time"
    # TYPE_UUID = "uuid" # TODO
    TYPE_URL = "url"

    TYPE_CHOICES = (
        (TYPE_BOOL, TYPE_BOOL),
        # (TYPE_COLOR, TYPE_COLOR, ),
        (TYPE_DATE, TYPE_DATE),
        (TYPE_DATETIME, TYPE_DATETIME),
        (TYPE_DECIMAL, TYPE_DECIMAL),
        (TYPE_DURATION, TYPE_DURATION),
        (TYPE_EMAIL, TYPE_EMAIL),
        (TYPE_FILE, TYPE_FILE),
        (TYPE_FLOAT, TYPE_FLOAT),
        # (TYPE_HTML, TYPE_HTML, ),
        (TYPE_IMAGE, TYPE_IMAGE),
        (TYPE_INT, TYPE_INT),
        (TYPE_JSON, TYPE_JSON),
        (TYPE_STRING, TYPE_STRING),
        (TYPE_TEXT, TYPE_TEXT),
        (TYPE_TIME, TYPE_TIME),
        # (TYPE_UUID, TYPE_UUID, ),
        (TYPE_URL, TYPE_URL),
    )

    name = models.CharField(
        max_length=255,
        unique=True,
        verbose_name=_("Name"),
        help_text=_("(e.g. SETTING_NAME)"),
    )
    value_type = models.CharField(
        max_length=20,
        choices=TYPE_CHOICES,
        verbose_name=_(
            "Type",
        ),
    )
    description = models.TextField(
        blank=True,
        null=True,
        verbose_name=_("Description"),
    )

    value_bool = models.BooleanField(
        default=False,
        verbose_name=_("Value"),
    )
    value_date = models.DateField(
        blank=True,
        null=True,
        verbose_name=_("Value"),
    )
    value_datetime = models.DateTimeField(
        blank=True,
        null=True,
        verbose_name=_("Value"),
    )
    value_decimal = models.DecimalField(
        blank=True,
        max_digits=19,
        decimal_places=10,
        default=Decimal("0.0"),
        verbose_name=_("Value"),
    )
    value_duration = models.DurationField(
        blank=True,
        null=True,
        verbose_name=_("Value"),
    )
    value_email = models.EmailField(
        blank=True,
        verbose_name=_("Value"),
    )
    value_file = models.FileField(
        blank=True,
        upload_to=fields.upload_to_files,
        verbose_name=_("Value"),
    )
    value_float = models.FloatField(
        blank=True,
        default=0.0,
        verbose_name=_("Value"),
    )
    value_image = models.FileField(
        blank=True,
        upload_to=fields.upload_to_images,
        verbose_name=_("Value"),
    )
    value_int = models.IntegerField(
        blank=True,
        default=0,
        verbose_name=_("Value"),
    )
    value_json = JSONField(
        blank=True,
        default=dict,
        verbose_name=_("Value"),
    )
    value_string = models.CharField(
        blank=True,
        max_length=255,
        verbose_name=_("Value"),
    )
    value_text = models.TextField(
        blank=True,
        verbose_name=_("Value"),
    )
    value_time = models.TimeField(
        blank=True,
        null=True,
        verbose_name=_("Value"),
    )
    value_url = models.URLField(
        blank=True,
        verbose_name=_("Value"),
    )
    validator = models.CharField(
        blank=True,
        null=True,
        max_length=255,
        verbose_name=_("Validator"),
        help_text=_(
            "Full python path to a validator function, "
            "eg. 'myapp.mymodule.my_validator'"
        ),
    )

    @property
    def value_field_name(self):
        return f"value_{self.value_type}"

    @property
    def value(self):
        return getattr(self, self.value_field_name, None)

    @value.setter
    def value(self, new_value):
        setattr(self, self.value_field_name, new_value)

    def __init__(self, *args, **kwargs):
        value = kwargs.pop("value", None)
        super().__init__(*args, **kwargs)
        if value is not None:
            self.value = value
        self.name_initial = self.name

    def validate(self):
        if not self.validator:
            return
        validator_func = import_function(self.validator)
        if not validator_func:
            raise ValueError(f"Invalid validator function path: {self.validator!r}")
        is_valid = validator_func(self.value)
        if not is_valid:
            raise ValidationError(
                f"Invalid value for setting {self.name!r}, "
                f"{self.value!r} is not validated by {self.validator!r} validator."
            )

    def save(self, *args, **kwargs):
        if settings.EXTRA_SETTINGS_ENFORCE_UPPERCASE_SETTINGS:
            self.name = enforce_uppercase_setting(self.name)
        super().save(*args, **kwargs)

    def clean(self):
        super().clean()
        self.validate()

    class Meta:
        ordering = ["name"]
        verbose_name = _("Setting")
        verbose_name_plural = _("Settings")

    def __str__(self):
        return force_str(f"{self.name} [{self.value_type}]")



__author__ = "Fabio Caccamo"
__copyright__ = "Copyright (c) 2020-present Fabio Caccamo"
__description__ = "config and manage typed extra settings using just the django admin."
__email__ = "fabio.caccamo@gmail.com"
__license__ = "MIT"
__title__ = "django-extra-settings"
__version__ = "0.12.0"



from django.db import migrations, models

import extra_settings.fields


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0002_auto_20200826_1714"),
    ]

    operations = [
        migrations.AlterField(
            model_name="setting",
            name="value_file",
            field=models.FileField(
                blank=True,
                upload_to=extra_settings.fields.upload_to_files,
                verbose_name="Value",
            ),
        ),
        migrations.AlterField(
            model_name="setting",
            name="value_image",
            field=models.FileField(
                blank=True,
                upload_to=extra_settings.fields.upload_to_images,
                verbose_name="Value",
            ),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="setting",
            name="value_duration",
            field=models.DurationField(blank=True, null=True, verbose_name="Value"),
        ),
        migrations.AlterField(
            model_name="setting",
            name="value_type",
            field=models.CharField(
                choices=[
                    ("bool", "bool"),
                    ("date", "date"),
                    ("datetime", "datetime"),
                    ("decimal", "decimal"),
                    ("duration", "duration"),
                    ("email", "email"),
                    ("file", "file"),
                    ("float", "float"),
                    ("image", "image"),
                    ("int", "int"),
                    ("string", "string"),
                    ("text", "text"),
                    ("time", "time"),
                    ("url", "url"),
                ],
                max_length=20,
                verbose_name="Type",
            ),
        ),
    ]



from django import forms
from django.conf import settings

from extra_settings.models import Setting
from extra_settings.utils import enforce_uppercase_setting


class SettingForm(forms.ModelForm):
    class Meta:
        model = Setting
        fields = "__all__"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # resize text area
        if "description" in self.fields:
            self.fields["description"].widget = forms.Textarea(
                attrs={"rows": 3, "cols": 51}
            )
        if "value_text" in self.fields:
            self.fields["value_text"].widget = forms.Textarea(
                attrs={"rows": 5, "cols": 51}
            )

    def clean_name(self):
        value = self.cleaned_data.get("name", "")
        if settings.EXTRA_SETTINGS_ENFORCE_UPPERCASE_SETTINGS:
            value = enforce_uppercase_setting(value)
        if hasattr(settings, value):
            raise forms.ValidationError(
                f"Invalid setting name, settings.{value} already "
                "defined in django.conf.settings."
            )
        return value



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0006_setting_validator"),
    ]

    operations = [
        migrations.AlterField(
            model_name="setting",
            name="validator",
            field=models.CharField(
                blank=True,
                help_text=(
                    "Full python path to a validator function, "
                    "eg. 'myapp.mymodule.my_validator'"
                ),
                max_length=255,
                null=True,
                verbose_name="Validator",
            ),
        ),
    ]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0005_setting_value_json_alter_setting_value_type"),
    ]

    operations = [
        migrations.AddField(
            model_name="setting",
            name="validator",
            field=models.CharField(
                blank=True, null=True, max_length=255, verbose_name="Validator"
            ),
        ),
    ]



from django.test import TestCase

from extra_settings.cache import (
    del_cached_setting,
    get_cached_setting,
    set_cached_setting,
)


class ExtraSettingsCacheTestCase(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_cache_del(self):
        set_cached_setting("test_cache_del", "del")
        val = get_cached_setting("test_cache_del")
        self.assertEqual(val, "del")
        del_cached_setting("test_cache_del")
        val = get_cached_setting("test_cache_del")
        self.assertEqual(val, None)

    def test_cache_get(self):
        val = get_cached_setting("test_cache_get")
        self.assertEqual(val, None)
        set_cached_setting("test_cache_get", "get")
        val = get_cached_setting("test_cache_get")
        self.assertEqual(val, "get")

    def test_cache_set(self):
        set_cached_setting("test_cache_set", "set")
        val = get_cached_setting("test_cache_set")
        self.assertEqual(val, "set")



from django.core.exceptions import ImproperlyConfigured

from extra_settings.metadata import (
    __author__,
    __copyright__,
    __description__,
    __license__,
    __title__,
    __version__,
)

try:
    from extra_settings import settings  # noqa: F401
except ImproperlyConfigured:
    pass

default_app_config = "extra_settings.apps.ExtraSettingsConfig"

__all__ = [
    "__author__",
    "__copyright__",
    "__description__",
    "__license__",
    "__title__",
    "__version__",
]



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0003_modified_upload_to"),
    ]

    operations = [
        migrations.AddField(
            model_name="setting",
            name="description",
            field=models.TextField(blank=True, null=True, verbose_name="Description"),
        ),
    ]



import re

from django.test import TestCase

from extra_settings.metadata import (
    __author__,
    __copyright__,
    __description__,
    __email__,
    __license__,
    __title__,
    __version__,
)


class MetadataTestCase(TestCase):
    """
    This class describes a metadata test case.
    """

    def test_metadata(self):
        self.assertTrue(isinstance(__author__, str))
        self.assertTrue(isinstance(__copyright__, str))
        self.assertTrue(isinstance(__description__, str))
        self.assertTrue(isinstance(__email__, str))
        self.assertTrue(isinstance(__license__, str))
        self.assertTrue(isinstance(__title__, str))
        self.assertTrue(isinstance(__version__, str))

    def test_version(self):
        v = __version__
        v_re = re.compile(r"^([0-9]+)(\.([0-9]+)){1,2}$")
        v_match = v_re.match(v)
        self.assertTrue(v_match is not None)



# Generated by Django 3.2.8 on 2022-10-24 07:17

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("extra_settings", "0007_alter_settings_validator"),
    ]

    operations = [
        migrations.AlterField(
            model_name="setting",
            name="name",
            field=models.CharField(
                help_text="(e.g. SETTING_NAME)",
                max_length=255,
                unique=True,
                verbose_name="Name",
            ),
        ),
        migrations.AlterField(
            model_name="setting",
            name="value_string",
            field=models.CharField(blank=True, max_length=255, verbose_name="Value"),
        ),
    ]



from inspect import isclass

from django.conf import settings
from django.db.models.signals import post_delete, post_migrate, post_save

from extra_settings.cache import del_cached_setting, set_cached_setting
from extra_settings.models import Setting


def __is_setting_signal(sender):
    return isclass(sender) and issubclass(sender, Setting)


def post_migrate_setting(sender, **kwargs):
    if not __is_setting_signal(sender):
        return
    Setting.set_defaults(settings.EXTRA_SETTINGS_DEFAULTS)


def post_save_setting(sender, instance, **kwargs):
    if not __is_setting_signal(sender):
        return
    if instance.name != instance.name_initial:
        del_cached_setting(instance.name_initial)
    set_cached_setting(instance.name, instance.value)


def post_delete_setting(sender, instance, **kwargs):
    if not __is_setting_signal(sender):
        return
    del_cached_setting(instance.name)


post_migrate.connect(
    post_migrate_setting,
    dispatch_uid="post_migrate_extra_settings_setting",
)
post_save.connect(
    post_save_setting,
    dispatch_uid="post_save_extra_settings_setting",
)
post_delete.connect(
    post_delete_setting,
    dispatch_uid="post_delete_extra_settings_setting",
)



from django.apps import AppConfig
from django.conf import settings
from django.db.models.signals import post_migrate


class ExtraSettingsConfig(AppConfig):
    name = "extra_settings"
    verbose_name = settings.EXTRA_SETTINGS_VERBOSE_NAME
    default_auto_field = "django.db.models.AutoField"

    def ready(self):
        from extra_settings import signals  # noqa: F401
        from extra_settings.models import Setting

        post_migrate.connect(Setting.set_defaults_from_settings, sender=self)



from django.template import Context, Template
from django.test import TestCase

from extra_settings.cache import del_cached_setting
from extra_settings.models import Setting


class ExtraSettingsTemplateTagsTestCase(TestCase):
    def setUp(self):
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="PACKAGE_NAME",
            defaults={
                "value_type": Setting.TYPE_STRING,
                "value_string": "django-extra-settings",
            },
        )

    def tearDown(self):
        pass

    def _render_template(self, string, context=None):
        return Template(string).render(Context(context or {}))

    def test_get_setting(self):
        rendered = self._render_template(
            "{% load extra_settings %}" '{% get_setting "PACKAGE_NAME" %}'
        )
        self.assertEqual(rendered, "django-extra-settings")

    def test_get_setting_assignment(self):
        rendered = self._render_template(
            "{% load extra_settings %}"
            '{% get_setting "PACKAGE_NAME" as package_name %}'
            "{{ package_name }}"
        )
        self.assertEqual(rendered, "django-extra-settings")

    def test_get_setting_default(self):
        rendered = self._render_template(
            "{% load extra_settings %}"
            '{% get_setting "INVALID_SETTING_NAME" default="ok" %}'
        )
        self.assertEqual(rendered, "ok")

    def test_get_setting_with_fallback_auto(self):
        rendered = self._render_template(
            "{% load extra_settings %}"
            '{% get_setting "EXTRA_SETTINGS_TEST_FALLBACK_VALUE" %}'
        )
        self.assertEqual(rendered, "fallback-value")

    def test_get_setting_with_fallback(self):
        del_cached_setting("EXTRA_SETTINGS_TEST_FALLBACK_VALUE")
        with self.settings(EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS=True):
            rendered = self._render_template(
                "{% load extra_settings %}"
                '{% get_setting "EXTRA_SETTINGS_TEST_FALLBACK_VALUE" %}'
            )
            self.assertEqual(rendered, "fallback-value")

    def test_get_setting_without_fallback(self):
        del_cached_setting("EXTRA_SETTINGS_TEST_FALLBACK_VALUE")
        with self.settings(EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS=False):
            rendered = self._render_template(
                "{% load extra_settings %}"
                '{% get_setting "EXTRA_SETTINGS_TEST_FALLBACK_VALUE" %}'
            )
            self.assertEqual(rendered, "")

    def test_get_setting_without_fallback_default(self):
        del_cached_setting("EXTRA_SETTINGS_TEST_FALLBACK_VALUE")
        with self.settings(EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS=False):
            rendered = self._render_template(
                "{% load extra_settings %}"
                "{% get_setting 'EXTRA_SETTINGS_TEST_FALLBACK_VALUE' default='fallback-default' %}"  # noqa: B950, E501
            )
            self.assertEqual(rendered, "fallback-default")



from django.test import TestCase

from extra_settings.models import Setting
from extra_settings.test import override_settings


class ExtraSettingsOverrideSettingTestCase(TestCase):
    def setUp(self):
        Setting.objects.bulk_create(
            [
                Setting(
                    name="TEST_OVERRIDE_SETTING_BOOL",
                    value_type=Setting.TYPE_BOOL,
                    value=True,
                ),
                Setting(
                    name="TEST_OVERRIDE_SETTING_STRING",
                    value_type=Setting.TYPE_STRING,
                    value="not overridden...",
                ),
            ]
        )

    def tearDown(self):
        pass

    def test_without_override_settings(self):
        value = Setting.get("TEST_OVERRIDE_SETTING_BOOL")
        self.assertEqual(value, True)
        value = Setting.get("TEST_OVERRIDE_SETTING_STRING")
        self.assertEqual(value, "not overridden...")

    @override_settings(
        TEST_OVERRIDE_SETTING_BOOL=False,
        TEST_OVERRIDE_SETTING_STRING="overridden!",
    )
    def test_with_override_settings_decorator(self):
        value = Setting.get("TEST_OVERRIDE_SETTING_BOOL")
        self.assertEqual(value, False)
        value = Setting.get("TEST_OVERRIDE_SETTING_STRING")
        self.assertEqual(value, "overridden!")

    def test_with_override_settings_context_manager(self):
        with override_settings(
            TEST_OVERRIDE_SETTING_BOOL=False,
            TEST_OVERRIDE_SETTING_STRING="overridden!",
        ):
            value = Setting.get("TEST_OVERRIDE_SETTING_BOOL")
            self.assertEqual(value, False)
            value = Setting.get("TEST_OVERRIDE_SETTING_STRING")
            self.assertEqual(value, "overridden!")



import os

from django.conf import settings


def upload_to_files(obj, filename):
    """Provide upload path for a file."""
    return _upload_to(settings.EXTRA_SETTINGS_FILE_UPLOAD_TO, filename)


def upload_to_images(obj, filename):
    """Provide upload path for an image."""
    return _upload_to(settings.EXTRA_SETTINGS_IMAGE_UPLOAD_TO, filename)


def _upload_to(directory, filename):
    """
    Build a path from given parameters and joins them.

    :param directory: A directory for upload.
    :param filename: The filename is to be saved.
    :return: Complete upload path under the django MEDIA_ROOT.
    """
    return os.path.join(directory, filename)



from django.conf import settings
from django.contrib import admin
from django.contrib.admin.sites import NotRegistered
from django.core.exceptions import ImproperlyConfigured
from django.utils.translation import gettext_lazy as _

from extra_settings.forms import SettingForm
from extra_settings.models import Setting


class SettingNamePrefixFilter(admin.SimpleListFilter):
    title = _("Name Prefix")
    parameter_name = "name_prefix"

    def lookups(self, request, model_admin):
        sep = "_"
        names = list(set(Setting.objects.all().values_list("name", flat=True)))
        names_count = len(names)
        names_parts = [name.split(sep) for name in names]
        # generate all prefixes
        prefixes = {}
        for name_parts in names_parts:
            name_parts_steps = []
            for name_part in name_parts:
                name_parts_steps.append(name_part)
                prefix = sep.join(name_parts_steps)
                prefixes.setdefault(prefix, 0)
                prefixes[prefix] += 1
        # keep only prefixes that match more than one entry and less than all entries
        prefixes = {
            key: value
            for key, value in prefixes.items()
            if value > 1 and value < names_count
        }
        # remove parent names with the same number of children of a child
        names = set(prefixes.keys())
        for name in names:
            for other_name in names:
                if (
                    name != other_name
                    and name.startswith(other_name)
                    and prefixes.get(name) == prefixes.get(other_name)
                ):
                    prefixes.pop(other_name, None)
        # sort prefixes alphabetically
        names = sorted(prefixes.keys())
        return [(name, f"{name} ({prefixes[name]})") for name in names]

    def queryset(self, request, queryset):
        prefix = self.value()
        if prefix:
            return queryset.filter(name__istartswith=prefix)
        return queryset


class SettingAdmin(admin.ModelAdmin):
    form = SettingForm
    value_fields_names = (
        "value_bool",
        "value_date",
        "value_datetime",
        "value_decimal",
        "value_duration",
        "value_email",
        "value_file",
        "value_float",
        "value_image",
        "value_int",
        "value_json",
        "value_string",
        "value_text",
        "value_time",
        "value_url",
    )
    search_fields = ("name",)
    list_display = ("name", "value_type") + value_fields_names + ("description",)

    # begin dynamic list filters

    list_filter_items = []

    if settings.EXTRA_SETTINGS_SHOW_NAME_PREFIX_LIST_FILTER:
        list_filter_items.append(SettingNamePrefixFilter)

    if settings.EXTRA_SETTINGS_SHOW_TYPE_LIST_FILTER:
        list_filter_items.append("value_type")

    if list_filter_items:
        list_filter = tuple(list_filter_items)

    # end dynamic list filters

    list_editable = value_fields_names
    sortable_by = ("name",)

    def get_changelist_form(self, request, **kwargs):
        return SettingForm

    def get_fieldsets(self, request, obj=None):
        if obj:
            fields = (
                "name",
                "value_type",
                obj.value_field_name,
                "validator",
                "description",
            )
        else:
            fields = (
                "name",
                "value_type",
            )
        return (
            (
                None,
                {
                    "classes": ("wide",),
                    "fields": fields,
                },
            ),
        )

    def get_readonly_fields(self, request, obj=None):
        return ("value_type",) if obj else ()

    class Media:
        css = {
            "all": ("extra_settings/css/extra_settings.css",),
        }
        js = ("extra_settings/js/extra_settings.js",)


def register_extra_settings_admin(
    app,
    queryset_processor=None,
    unregister_default=True,
):
    """
    Based on django dynamic models general-purpose approach.
    https://code.djangoproject.com/wiki/DynamicModels#Ageneral-purposeapproach
    """
    apps = settings.INSTALLED_APPS
    if app not in apps:
        if "." in app:
            app = app.split(".")[0]
        if app not in apps:
            raise ImproperlyConfigured(
                f"{app!r} application not listed in settings.INSTALLED_APPS."
            )

    # create dynamic proxy model
    class Meta:
        proxy = True

    attrs = {
        "__module__": app,
        "Meta": Meta,
    }

    SettingProxyModel = type("Setting", (Setting,), attrs)

    # create dynamic model admin
    @admin.register(SettingProxyModel)
    class SettingProxyAdmin(SettingAdmin):
        queryset_processor = None

        def get_queryset(self, request):
            qs = super().get_queryset(request)
            if self.queryset_processor and callable(self.queryset_processor):
                qs = queryset_processor(qs)
            return qs

    SettingProxyAdmin.queryset_processor = queryset_processor

    # register dynamic model admin and unregister default model admin

    if unregister_default:
        try:
            admin.site.unregister(Setting)
        except NotRegistered:
            pass


admin.site.register(Setting, SettingAdmin)


app = settings.EXTRA_SETTINGS_ADMIN_APP
if app and app != "extra_settings":
    register_extra_settings_admin(
        app=app,
        queryset_processor=None,
        unregister_default=True,
    )



from django.contrib.admin.sites import AdminSite
from django.test import TestCase

from extra_settings.admin import SettingAdmin
from extra_settings.forms import SettingForm
from extra_settings.models import Setting


class MockRequest:
    pass


class MockSuperUser:
    def has_perm(self, perm):
        return True


request = MockRequest()
request.user = MockSuperUser()


class ExtraSettingsAdminTestCase(TestCase):
    def setUp(self):
        self._setting_obj, setting_created = Setting.objects.get_or_create(
            name="PACKAGE_NAME",
            defaults={
                "value_type": Setting.TYPE_STRING,
                "value_string": "django-extra-settings",
            },
        )
        self._site = AdminSite()

    def tearDown(self):
        pass

    def test_changelist_form(self):
        ma = SettingAdmin(model=Setting, admin_site=AdminSite())
        self.assertEqual(ma.get_changelist_form(request=None), SettingForm)

    def test_fieldsets(self):
        ma = SettingAdmin(model=Setting, admin_site=AdminSite())
        self.assertEqual(
            ma.get_fieldsets(request),
            (
                (
                    None,
                    {
                        "classes": ("wide",),
                        "fields": ("name", "value_type"),
                    },
                ),
            ),
        )
        self.assertEqual(
            ma.get_fieldsets(request, self._setting_obj),
            (
                (
                    None,
                    {
                        "classes": ("wide",),
                        "fields": (
                            "name",
                            "value_type",
                            "value_string",
                            "validator",
                            "description",
                        ),
                    },
                ),
            ),
        )

    def test_modeladmin_save(self):
        ma = SettingAdmin(model=Setting, admin_site=AdminSite())
        ma.save_model(obj=Setting(), request=None, form=None, change=None)

    def test_modeladmin_str(self):
        ma = SettingAdmin(model=Setting, admin_site=AdminSite())
        self.assertEqual(str(ma), "extra_settings.SettingAdmin")

    def test_readonly_fields(self):
        ma = SettingAdmin(model=Setting, admin_site=AdminSite())
        self.assertEqual(ma.get_readonly_fields(request), ())
        self.assertEqual(
            ma.get_readonly_fields(request, self._setting_obj), ("value_type",)
        )



from django.test import TestCase

from extra_settings.forms import SettingForm
from extra_settings.models import Setting


class ExtraSettingsFormsTestCase(TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_form_with_valid_data(self):
        form_data = {"name": "PACKAGE_NAME", "value_type": Setting.TYPE_BOOL}
        form_obj = SettingForm(data=form_data)
        self.assertTrue(form_obj.is_valid())

    def test_form_with_incomplete_data(self):
        form_data = {"name": "PACKAGE_NAME"}
        form_obj = SettingForm(data=form_data)
        self.assertFalse(form_obj.is_valid())

    def test_form_with_invalid_setting_name(self):
        form_data = {"name": "INSTALLED_APPS", "value_type": Setting.TYPE_BOOL}
        form_obj = SettingForm(data=form_data)
        self.assertFalse(form_obj.is_valid())

    def test_form_with_optional_data(self):
        form_data = {
            "name": "PACKAGE_NAME",
            "value_type": Setting.TYPE_BOOL,
            "description": "Yes/No",
        }
        form_obj = SettingForm(data=form_data)
        self.assertTrue(form_obj.is_valid())



from decimal import Decimal

from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name="Setting",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                (
                    "name",
                    models.CharField(
                        help_text="(e.g. SETTING_NAME)",
                        max_length=50,
                        unique=True,
                        verbose_name="Name",
                    ),
                ),
                (
                    "value_type",
                    models.CharField(
                        choices=[
                            ("bool", "bool"),
                            ("date", "date"),
                            ("datetime", "datetime"),
                            ("decimal", "decimal"),
                            ("email", "email"),
                            ("file", "file"),
                            ("float", "float"),
                            ("image", "image"),
                            ("int", "int"),
                            ("string", "string"),
                            ("text", "text"),
                            ("time", "time"),
                            ("url", "url"),
                        ],
                        max_length=20,
                        verbose_name="Type",
                    ),
                ),
                (
                    "value_bool",
                    models.BooleanField(default=False, verbose_name="Value"),
                ),
                (
                    "value_date",
                    models.DateField(blank=True, null=True, verbose_name="Value"),
                ),
                (
                    "value_datetime",
                    models.DateTimeField(blank=True, null=True, verbose_name="Value"),
                ),
                (
                    "value_decimal",
                    models.DecimalField(
                        blank=True,
                        decimal_places=10,
                        default=Decimal("0.0"),
                        max_digits=19,
                        verbose_name="Value",
                    ),
                ),
                (
                    "value_email",
                    models.EmailField(blank=True, max_length=254, verbose_name="Value"),
                ),
                (
                    "value_file",
                    models.FileField(
                        blank=True, upload_to="files", verbose_name="Value"
                    ),
                ),
                (
                    "value_float",
                    models.FloatField(blank=True, default=0.0, verbose_name="Value"),
                ),
                (
                    "value_image",
                    models.FileField(
                        blank=True, upload_to="images", verbose_name="Value"
                    ),
                ),
                (
                    "value_int",
                    models.IntegerField(blank=True, default=0, verbose_name="Value"),
                ),
                (
                    "value_string",
                    models.CharField(blank=True, max_length=50, verbose_name="Value"),
                ),
                (
                    "value_text",
                    models.TextField(blank=True, verbose_name="Value"),
                ),
                (
                    "value_time",
                    models.TimeField(blank=True, null=True, verbose_name="Value"),
                ),
                (
                    "value_url",
                    models.URLField(blank=True, verbose_name="Value"),
                ),
            ],
            options={
                "verbose_name": "Setting",
                "verbose_name_plural": "Settings",
                "ordering": ["name"],
            },
        ),
    ]



from django.core.exceptions import ValidationError
from django.test import TestCase

from extra_settings.models import Setting


def positive_int_validator(value):
    return value > 0


def alphanumeric_strings_validator(value):
    return value.isalnum()


class ExtraSettingsValidatorsTestCase(TestCase):
    def setUp(self):
        Setting.objects.bulk_create(
            [
                Setting(
                    name="TEST_VALIDATE_POSITIVE_INTEGER",
                    value_type=Setting.TYPE_INT,
                    value=10,
                    validator="tests.test_validators.positive_int_validator",
                ),
                Setting(
                    name="TEST_NO_VALIDATORS_INTEGER",
                    value_type=Setting.TYPE_INT,
                    value=10,
                ),
                Setting(
                    name="TEST_VALIDATE_STRINGS",
                    value_type=Setting.TYPE_STRING,
                    value="This is a correct String",
                    validator="tests.test_validators.alphanumeric_strings_validator",
                ),
                Setting(
                    name="TEST_VALIDATE_WITH_INVALID_VALIDATOR",
                    value_type=Setting.TYPE_STRING,
                    value="This is a correct String",
                    validator="tests.test_validators.invalid_validator",
                ),
            ]
        )

    def test_validators(self):
        positive_integer = Setting.objects.get(name="TEST_VALIDATE_POSITIVE_INTEGER")
        positive_integer.value = -10
        with self.assertRaises(ValidationError):
            positive_integer.full_clean()
        normal_integer = Setting.objects.get(name="TEST_NO_VALIDATORS_INTEGER")
        normal_integer.value = -10
        normal_integer.full_clean()
        alnum_string = Setting.objects.get(name="TEST_VALIDATE_STRINGS")
        alnum_string.value = "!@-10"
        with self.assertRaises(ValidationError):
            alnum_string.full_clean()

    def test_invalid_validator(self):
        setting_obj = Setting.objects.get(name="TEST_VALIDATE_WITH_INVALID_VALIDATOR")
        with self.assertRaises(ValueError):
            setting_obj.full_clean()



from extra_settings.cache import del_cached_setting, set_cached_setting
from extra_settings.models import Setting


class override_settings:
    """
    This class describes the override_settings decorator / context manager.
    """

    def __init__(self, **settings):
        self._settings = {}
        self._settings_override = settings.copy()

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            self._patch_settings()
            func(*args, **kwargs)
            self._reset_settings()

        return wrapped

    def __enter__(self):
        self._patch_settings()
        return self

    def __exit__(self, e_type, e_value, e_traceback):
        self._reset_settings()

    def _patch_settings(self):
        for key, value in self._settings_override.items():
            self._settings[key] = Setting.get(key)
            set_cached_setting(key, value)

    def _reset_settings(self):
        for key, _ in self._settings_override.items():
            del_cached_setting(key)
        for key, value in self._settings.items():
            set_cached_setting(key, value)



from django.utils.module_loading import import_string
from django.utils.text import slugify


def enforce_uppercase_setting(name):
    return slugify(name).replace("-", "_").upper()


def import_function(path):
    if not path:
        return None
    try:
        func = import_string(path)
        if callable(func):
            return func
        return None
    except ImportError:
        return None



from django.conf import settings
from django.core.cache import cache, caches


def _get_cache():
    cache_name = settings.EXTRA_SETTINGS_CACHE_NAME
    return caches[cache_name] if cache_name in settings.CACHES else cache


def _get_cache_key(key):
    return f"extra_settings_{key}"


def del_cached_setting(key):
    _get_cache().delete(_get_cache_key(key))


def get_cached_setting(key):
    return _get_cache().get(_get_cache_key(key), None)


def set_cached_setting(key, value):
    _get_cache().set(_get_cache_key(key), value)



from django import template

from extra_settings.models import Setting

register = template.Library()


@register.simple_tag(takes_context=True)
def get_setting(context, name, default=""):
    return Setting.get(name, default)



from django.test import TestCase

from extra_settings.models import Setting


class ExtraSettingsModelsTestCase(TestCase):
    def setUp(self):
        Setting.objects.bulk_create(
            [
                Setting(
                    name="TEST_SETTING_BOOL",
                    value_type=Setting.TYPE_BOOL,
                ),
                # Setting(
                #     name="TEST_SETTING_COLOR",
                #     value_type=Setting.TYPE_COLOR,
                # ),
                Setting(
                    name="TEST_SETTING_DATE",
                    value_type=Setting.TYPE_DATE,
                ),
                Setting(
                    name="TEST_SETTING_DATETIME",
                    value_type=Setting.TYPE_DATETIME,
                ),
                Setting(
                    name="TEST_SETTING_DURATION",
                    value_type=Setting.TYPE_DURATION,
                ),
                Setting(
                    name="TEST_SETTING_DECIMAL",
                    value_type=Setting.TYPE_DECIMAL,
                ),
                Setting(
                    name="TEST_SETTING_EMAIL",
                    value_type=Setting.TYPE_EMAIL,
                ),
                Setting(
                    name="TEST_SETTING_FILE",
                    value_type=Setting.TYPE_FILE,
                ),
                Setting(
                    name="TEST_SETTING_FLOAT",
                    value_type=Setting.TYPE_FLOAT,
                ),
                # Setting(
                #     name="TEST_SETTING_HTML",
                #     value_type=Setting.TYPE_HTML,
                # ),
                Setting(
                    name="TEST_SETTING_IMAGE",
                    value_type=Setting.TYPE_IMAGE,
                ),
                Setting(
                    name="TEST_SETTING_INT",
                    value_type=Setting.TYPE_INT,
                ),
                Setting(
                    name="TEST_SETTING_JSON",
                    value_type=Setting.TYPE_JSON,
                ),
                Setting(
                    name="TEST_SETTING_STRING",
                    value_type=Setting.TYPE_STRING,
                ),
                Setting(
                    name="TEST_SETTING_TEXT",
                    value_type=Setting.TYPE_TEXT,
                ),
                Setting(
                    name="TEST_SETTING_TIME",
                    value_type=Setting.TYPE_TIME,
                ),
                # Setting(name='TEST_SETTING_UUID', value_type=Setting.TYPE_UUID),
                Setting(
                    name="TEST_SETTING_URL",
                    value_type=Setting.TYPE_URL,
                ),
            ]
        )

    def tearDown(self):
        pass

    def test_create_setting(self):
        # bool
        setting_value = True
        setting_obj = Setting(
            name="TEST_CREATE_DYNAMIC_SETTING_BOOL",
            value_type=Setting.TYPE_BOOL,
            value=setting_value,
        )
        setting_obj.save()
        self.assertEqual(setting_obj.value, setting_value)
        # url
        setting_value = "https://github.com/fabiocaccamo/django-extra-settings"
        setting_obj = Setting(
            name="TEST_CREATE_DYNAMIC_SETTING_URL",
            value_type=Setting.TYPE_URL,
            value=setting_value,
        )
        setting_obj.save()
        self.assertEqual(setting_obj.value, setting_value)

    def test_getter_setter(self):
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="TEST_GETTER_SETTER", defaults={"value_type": Setting.TYPE_STRING}
        )
        self.assertEqual(setting_obj.value, "")
        setting_obj.value = "string value"
        setting_obj.save()
        setting_obj = Setting.objects.get(name="TEST_GETTER_SETTER")
        self.assertEqual(setting_obj.value, "string value")

    def test_get_with_valid_name(self):
        setting_string_value = Setting.get("TEST_SETTING_STRING")
        self.assertEqual(setting_string_value, "")

    def test_get_with_valid_name_and_default_value(self):
        setting_value = Setting.get(
            "TEST_SETTING_STRING", default="default string value"
        )
        self.assertEqual(setting_value, "")

    def test_get_with_invalid_name(self):
        setting_value = Setting.get("TEST_SETTING_STRING_INVALID")
        self.assertEqual(setting_value, None)

    def test_get_with_invalid_name_and_default_value(self):
        setting_value = Setting.get(
            "TEST_SETTING_STRING_INVALID", default="default string value"
        )
        self.assertEqual(setting_value, "default string value")

    def test_get_num_queries(self):
        with self.assertNumQueries(2):
            # first query get the setting
            # second query cache all settings
            # (executed only if the first one find a result)
            Setting.get("TEST_SETTING_STRING")
        with self.assertNumQueries(0):
            Setting.get("TEST_SETTING_STRING")
        with self.assertNumQueries(0):
            Setting.get("TEST_SETTING_STRING")

    def test_get_from_conf_settings_num_queries(self):
        with self.assertNumQueries(1):
            setting_value = Setting.get("EXTRA_SETTINGS_TEST_VALUE")
            self.assertEqual(setting_value, "value")
        with self.assertNumQueries(0):
            setting_value = Setting.get("EXTRA_SETTINGS_TEST_VALUE")
            self.assertEqual(setting_value, "value")

    def test_cache_updated_on_model_delete(self):
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="TEST_GETTER_SETTER", defaults={"value_type": Setting.TYPE_STRING}
        )
        setting_obj.value = "string value"
        setting_obj.save()
        self.assertEqual(Setting.get("TEST_GETTER_SETTER"), "string value")
        Setting.objects.filter(pk=setting_obj.pk).delete()
        self.assertEqual(Setting.get("TEST_GETTER_SETTER"), None)

    def test_cache_updated_on_model_name_changed(self):
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="TEST_GETTER_SETTER", defaults={"value_type": Setting.TYPE_STRING}
        )
        setting_obj.value = "string value"
        setting_obj.save()
        self.assertEqual(Setting.get("TEST_GETTER_SETTER"), "string value")
        setting_obj.name = "TEST_GETTER_SETTER_RENAMED"
        setting_obj.save()
        self.assertEqual(Setting.get("TEST_GETTER_SETTER"), None)
        self.assertEqual(Setting.get("TEST_GETTER_SETTER_RENAMED"), "string value")

    def test_repr(self):
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="PACKAGE_NAME",
            defaults={
                "value_type": Setting.TYPE_STRING,
                "value_string": "django-extra-settings",
            },
        )
        setting_repr = f"{setting_obj.name} [{setting_obj.value_type}]"
        self.assertEqual(f"{setting_obj}", setting_repr)

    def test_set_defaults_from_settings(self):
        self.assertEqual(
            Setting.get("TEST_DEFAULT_URL"),
            "https://github.com/fabiocaccamo/django-extra-settings",
        )

    def test_set_defaults(self):
        Setting.set_defaults([])
        defaults = [
            {
                "name": "TEST_DEFAULT_STRING",
                "type": "string",
                "value": "ok",
            },
            {
                "name": "TEST_DEFAULT_INT",
                "type": "int",
                "value": 2,
            },
        ]
        Setting.set_defaults(defaults)
        self.assertEqual(Setting.get("TEST_DEFAULT_STRING"), "ok")
        self.assertEqual(Setting.get("TEST_DEFAULT_INT"), 2)

    def test_set_defaults_with_invalid_setting(self):
        defaults = {
            "TEST_DEFAULT_STRING": {
                "name": "TEST_DEFAULT_STRING",
                "type": "string",
                "value": "ok",
            },
            "TEST_DEFAULT_INT": {
                "name": "TEST_DEFAULT_INT",
                "type": "int",
                "value": 2,
            },
        }
        with self.assertRaises(ValueError):
            Setting.set_defaults(defaults)
        defaults = [
            {
                "name": "TEST_DEFAULT_STRING",
                "value": "ok",
            },
            {
                "name": "TEST_DEFAULT_INT",
                "value": 2,
            },
        ]
        with self.assertRaises(ValueError):
            Setting.set_defaults(defaults)

    def test_setting_type_json(self):
        # getter & setter
        setting_obj, setting_created = Setting.objects.get_or_create(
            name="TEST_SETTING_JSON", defaults={"value_type": Setting.TYPE_JSON}
        )
        self.assertEqual(setting_obj.value, {})
        setting_obj.value = {"level": "L2", "role": "Admin"}
        setting_obj.save()
        setting_obj = Setting.objects.get(name="TEST_SETTING_JSON")
        self.assertEqual(setting_obj.value, {"level": "L2", "role": "Admin"})



from django.conf import settings

if not hasattr(settings, "EXTRA_SETTINGS_ADMIN_APP"):
    settings.EXTRA_SETTINGS_ADMIN_APP = "extra_settings"

if not hasattr(settings, "EXTRA_SETTINGS_CACHE_NAME"):
    settings.EXTRA_SETTINGS_CACHE_NAME = "extra_settings"

if not hasattr(settings, "EXTRA_SETTINGS_DEFAULTS"):
    settings.EXTRA_SETTINGS_DEFAULTS = []

if not hasattr(settings, "EXTRA_SETTINGS_ENFORCE_UPPERCASE_SETTINGS"):
    settings.EXTRA_SETTINGS_ENFORCE_UPPERCASE_SETTINGS = True

if not hasattr(settings, "EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS"):
    settings.EXTRA_SETTINGS_FALLBACK_TO_CONF_SETTINGS = True

if not hasattr(settings, "EXTRA_SETTINGS_FILE_UPLOAD_TO"):
    settings.EXTRA_SETTINGS_FILE_UPLOAD_TO = "files"

if not hasattr(settings, "EXTRA_SETTINGS_IMAGE_UPLOAD_TO"):
    settings.EXTRA_SETTINGS_IMAGE_UPLOAD_TO = "images"

if not hasattr(settings, "EXTRA_SETTINGS_SHOW_NAME_PREFIX_LIST_FILTER"):
    settings.EXTRA_SETTINGS_SHOW_NAME_PREFIX_LIST_FILTER = False

if not hasattr(settings, "EXTRA_SETTINGS_SHOW_TYPE_LIST_FILTER"):
    settings.EXTRA_SETTINGS_SHOW_TYPE_LIST_FILTER = False

if not hasattr(settings, "EXTRA_SETTINGS_VERBOSE_NAME"):
    settings.EXTRA_SETTINGS_VERBOSE_NAME = "Extra Settings"
